import random

def generate_matrix(rows, cols, mode='sequential', order='row', file_name='matrix.hex'):
    """
    Generate a 2D matrix and save it to a hex file.

    Parameters:
        rows (int): Number of rows
        cols (int): Number of columns
        mode (str): 'sequential', 'random', 'zeros', 'ones', or 'repeat_rows'
        order (str): 'row' or 'col' for row-major or column-major order
        file_name (str): Output file name
    """
    
    total = rows * cols

    if mode == 'sequential':
        values = list(range(total))
    elif mode == 'random':
        values = random.sample(range(0, 2**16), total)
    elif mode == 'zeros':
        values = [0] * total
    elif mode == 'ones':
        values = [1] * total
    elif mode == 'repeat_rows':
        # Create a column vector 1 to rows, repeat for each column
        matrix = [[i + 1 for _ in range(cols)] for i in range(rows)]
    else:
        raise ValueError("Mode must be 'sequential', 'random', 'zeros', 'ones', or 'repeat_rows'")

    # Only arrange the matrix if not in 'repeat_rows' mode
    if mode != 'repeat_rows':
        matrix = [[0] * cols for _ in range(rows)]

        if order == 'row':
            idx = 0
            for i in range(rows):
                for j in range(cols):
                    matrix[i][j] = values[idx]
                    idx += 1
        elif order == 'col':
            idx = 0
            for j in range(cols):
                for i in range(rows):
                    matrix[i][j] = values[idx]
                    idx += 1
        else:
            raise ValueError("Order must be 'row' or 'col'")

    # Write matrix to .hex file
    with open(file_name, 'w') as f:
        for i in range(rows):
            for j in range(cols):
                hex_value = format(matrix[i][j] & 0xFFFF, '04X')
                f.write(f"{hex_value}\n")

    print(f"Matrix ({rows}x{cols}) written to '{file_name}' in {mode}-{order if mode != 'repeat_rows' else 'col'} order.")


# Example usage:
if __name__ == "__main__":
    print("Welcome to the 2D Matrix Generator!")
    rows = int(input("Enter the number of rows: "))
    cols = int(input("Enter the number of columns: "))
    
    mode = input("Choose the mode ('sequential', 'random', 'zeros', 'ones', or 'repeat_rows'): ").strip().lower()
    order = input("Choose the order ('row' or 'col'): ").strip().lower() if mode != 'repeat_rows' else 'col'
    
    file_name = input("Enter the output file name (e.g., 'matrix.hex'): ").strip()

    generate_matrix(rows, cols, mode=mode, order=order, file_name=file_name)
