import numpy as np

# File paths
INPUT_FILE_PATH = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"
# OUTPUT_FILE_PATH = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"

# Memory and configuration parameters
MEM_ROW = 20
MEM_COL = 100
TOP_K = 4
DATA_WIDTH = 16

# A matrix size
size_of_A_row = 20
size_of_A_col = 20

A_reading_start_row = 0
A_reading_start_col = 52

# Start position of maskA, maskQ, maskKV in memory3 (output memory)
mask_reading_start_row = 0
mask_reading_start_col = 92

def read_memory(file_path):
    with open(file_path, 'r') as f:
        content = f.read().split()
    return np.array([int(val, 16) for val in content], dtype=np.uint16).reshape((MEM_ROW, MEM_COL))

def write_memory(file_path, memory):
    with open(file_path, 'w') as f:
        for row in memory:
            f.write(' '.join(f"{val:04x}" for val in row) + '\n')

# def extract_matrix_A(mem):
#     return mem[:size_of_A_row, :size_of_A_col].astype(np.int32)
def extract_matrix_A(mem):
    return mem[
        A_reading_start_row : A_reading_start_row + size_of_A_row,
        A_reading_start_col : A_reading_start_col + size_of_A_col
    ].astype(np.int32)

def compute_top_k_mask(A, k=TOP_K):
    mask_A = np.zeros_like(A, dtype=np.uint8)
    for i, row in enumerate(A):
        top_k_indices = np.argpartition(-row, k)[:k]
        mask_A[i, top_k_indices] = 1
    return mask_A

def pack_mask_rows(mask_A):
    packed = []
    for row in mask_A:
        bitstring = ''.join(str(b) for b in row)
        part1 = int(bitstring[:16], 2)
        part2 = int(bitstring[16:].ljust(16, '0'), 2)
        packed.append(part1)
        packed.append(part2)
    return np.array(packed, dtype=np.uint16)

# def unpack_mask_rows(packed, num_rows=20):
#     mask = []
#     for i in range(0, len(packed), 2):
#         bits = f"{packed[i]:016b}" + f"{packed[i+1]:016b}"
#         row_mask = [int(b) for b in bits[:20]]
#         mask.append(row_mask)
#     return np.array(mask, dtype=np.uint8)
def unpack_mask_rows(packed, num_rows=20, num_cols=20):
    mask = []
    for i in range(0, len(packed), 2):
        # print("packet[%d]= %x packet[%d+1]=%x" % (i, packed[i], i, packed[i+1]))#########
        # # value = (packed[i+1] << 16) #| packed[i]
        lsb = f"{packed[i]:016b}"      # 16 bits for lsb (e.g., 1020)
        msb = f"{packed[i+1]:04b}"     # 4 bits for msb (e.g., C)
        bitstring = msb + lsb          # Concatenate to get 20 bits
        # bitstring = f"{value:020b}"
        row_mask = [int(b) for b in bitstring[:num_cols]]
        # print("row_mask[%d]: %s" % (i//2, row_mask))####
        row_mask.reverse()  # Reverse the order of bits
        mask.append(row_mask)
        # print("row_mask_rev[%d]: %s" % (i//2, row_mask))####
    return np.array(mask, dtype=np.uint8)




def read_mask_A_from_memory_per_row(mem, start_row, start_col, rows=20):
    packed_values = []
    print("\nReading mask_A from memory per row (2 columns per row):")
    for r in range(start_row, start_row + rows):
        col1 = start_col
        col2 = start_col + 1
        val1 = mem[r, col1]
        val2 = mem[r, col2]
        # print(f"  Reading mask_A row {r-start_row} from memory[{r}][{col1}] = 0x{val1:04x}")
        # print(f"  Reading mask_A row {r-start_row} from memory[{r}][{col2}] = 0x{val2:04x}")
        packed_values.append(val1)
        packed_values.append(val2)
    return np.array(packed_values, dtype=np.uint16)

def read_mask_vector_from_memory(mem, start_row, start_col, length, label):
    values = []
    print(f"\nReading {label} from memory (1 column per row):")
    for r in range(start_row, start_row + length):
        val = mem[r, start_col]
        # print(f"  Reading {label}[{r - start_row}] from memory[{r}][{start_col}] = ")
        values.append(val)
    return np.array(values, dtype=np.uint16)

def validate_mask(expected, actual, name="mask"):
    if np.array_equal(expected, actual):
        print(f"\n✅ [PASS] {name} matches.\n")
    else:
        print(f"\n❌ [FAIL] {name} does not match.\n")
        print(f"Expected {name}:\n{expected}")
        print(f"Actual {name}:\n{actual}")

def main():
    # 1. Load memory (mem1.hex)
    memory = read_memory(INPUT_FILE_PATH)

    # 2. Extract matrix A
    A = extract_matrix_A(memory)
    print("Matrix A (20x20):")
    # print(A)
    print(np.vectorize(lambda x: f"0x{x:04x}")(A))

    # 3. Compute Top-K mask_A
    mask_A = compute_top_k_mask(A)
    print("\nComputed mask_A (Top-4 per row):")
    print(mask_A)

    # 4. Pack mask_A
    packed_mask_A = pack_mask_rows(mask_A)

    # 5. Read expected mask_A from memory3 (output memory layout)
    mem_mask_A = read_mask_A_from_memory_per_row(memory, mask_reading_start_row, mask_reading_start_col, rows=20)
    mem_mask_A_matrix = unpack_mask_rows(mem_mask_A)
    print("\nUnpacked mask_A read from memory:")
    print(mem_mask_A_matrix)

    # exit()
    # 6. Read mask_Q and mask_KV (each 20 elements, 1 column each)
    mask_Q_start_col = mask_reading_start_col + 2   # Assuming mask_Q stored right after mask_A (2 cols)
    mask_KV_start_col = mask_Q_start_col + 1       # mask_KV after mask_Q

    mem_mask_Q = read_mask_vector_from_memory(memory, mask_reading_start_row, mask_Q_start_col, 20, label="mask_Q")
    mem_mask_KV = read_mask_vector_from_memory(memory, mask_reading_start_row, mask_KV_start_col, 20, label="mask_KV")

    print("\nmask_Q (read):")
    print(mem_mask_Q)

    print("\nmask_KV (read):")
    print(mem_mask_KV)

    # 7. Validate mask_A
    validate_mask(mask_A, mem_mask_A_matrix, name="mask_A")

    # Optional: you can implement validation for mask_Q and mask_KV if you have expected results
    expected_mask_Q = np.array([1 if np.any(row) else 0 for row in mask_A], dtype=np.uint16)
    expected_mask_KV = np.array([1 if np.any(mask_A[:, col]) else 0 for col in range(mask_A.shape[1])], dtype=np.uint16)
    expected_mask_KV = expected_mask_KV[:20]  # Assuming 20 elements only (match mem layout)

    # 8. Validate mask_Q and mask_KV
    validate_mask(expected_mask_Q, mem_mask_Q, name="mask_Q")
    validate_mask(expected_mask_KV, mem_mask_KV, name="mask_KV")
    
    # 8. Write output memory (optional, writing computed packed mask_A back)
    output_memory = np.copy(memory)
    for r in range(mask_reading_start_row, mask_reading_start_row + 20):
        output_memory[r, mask_reading_start_col] = packed_mask_A[(r - mask_reading_start_row)*2]
        output_memory[r, mask_reading_start_col + 1] = packed_mask_A[(r - mask_reading_start_row)*2 + 1]

    # write_memory(OUTPUT_FILE_PATH, output_memory)
    # print(f"\nOutput memory written to {OUTPUT_FILE_PATH}")

if __name__ == "__main__":
    main()
