import os

# -------------------------------
# Helper Functions
# -------------------------------

def read_hex_file(file_path):
    with open(file_path, 'r') as f:
        lines = f.readlines()
    return [line.strip() for line in lines]

def write_hex_file(file_path, memory_data):
    with open(file_path, 'w') as f:
        for val in memory_data:
            f.write(val + '\n')

def get_index(row, col, total_cols):
    # Row-major order: each row contains 'total_cols' elements
    return row * total_cols + col

def extract_matrix(memory_data, start_row, start_col, rows, cols, total_mem_cols):
    matrix = []
    for r in range(rows):
        row_data = []
        for c in range(cols):
            idx = get_index(start_row + r, start_col + c, total_mem_cols)
            hex_val = memory_data[idx]
            row_data.append(int(hex_val, 16))
        matrix.append(row_data)
    return matrix

def insert_matrix_into_memory(memory_data, matrix, start_row, start_col, total_mem_cols):
    for r in range(len(matrix)):
        for c in range(len(matrix[0])):
            idx = get_index(start_row + r, start_col + c, total_mem_cols)
            memory_data[idx] = format(matrix[r][c], '04X')  # 4-digit uppercase hex

def multiply_matrices(A, B):
    result = []
    for i in range(len(A)):
        result_row = []
        for j in range(len(B[0])):
            sum = 0
            for k in range(len(B)):
                sum += A[i][k] * B[k][j]
            result_row.append(sum)
        result.append(result_row)
    return result

def compare_matrices(mat1, mat2):
    mismatch_found = False
    print("\n🔍 Comparing Computed Output vs Expected Output:\n")
    for i in range(len(mat1)):
        row_display = []
        for j in range(len(mat1[0])):
            val1 = mat1[i][j]
            val2 = mat2[i][j]
            if val1 != val2:
                mismatch_found = True
                cell = f"\033[91m{val1} != {val2} !!\033[0m"
            else:
                cell = f"{val1}"
            row_display.append(cell)
        print(f"Row {i:2d}: " + "\t".join(row_display))
    return not mismatch_found

def print_matrix(matrix, name):
    print(f"\n📦 Matrix: {name}")
    for i, row in enumerate(matrix):
        print(f"Row {i:2d}: " + "\t".join([f"{val:4d}" for val in row]))

# -------------------------------
# USER INPUT SECTION
# -------------------------------

# File and memory configuration
file_path = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"
mem_rows = 20       # number of matrix rows
mem_cols = 80       # number of matrix columns (each row has 80 hex values)

# Matrix A configuration
A_start_row = 0
A_start_col = 0
A_rows = 20
A_cols = 4

# Matrix B configuration
B_start_row = 0
B_start_col = 4#4
B_rows = 4
B_cols = 10

# Output Matrix (A x B) configuration
OUT_start_row = 0
OUT_start_col = 12#12

# -------------------------------
# MAIN PROGRAM
# -------------------------------

memory_data = read_hex_file(file_path)

# Extract matrices
A = extract_matrix(memory_data, A_start_row, A_start_col, A_rows, A_cols, mem_cols)
# B = extract_matrix(memory_data, B_start_row, B_start_col, B_rows, B_cols, mem_cols)
B_transposed = extract_matrix(memory_data, B_start_row, B_start_col, B_cols, B_rows, mem_cols)
B = list(map(list, zip(*B_transposed)))  # Transpose back to original B
Expected_Output = extract_matrix(memory_data, OUT_start_row, OUT_start_col, A_rows, B_cols, mem_cols)

# Compute matrix multiplication
Computed_Output = multiply_matrices(A, B)

# Print matrices
print_matrix(A, "Matrix A")
print_matrix(B, "Matrix B")
print_matrix(Computed_Output, "Computed Output")
print_matrix(Expected_Output, "Expected Output (from memory)")

# Compare and highlight mismatches
if compare_matrices(Computed_Output, Expected_Output):
    print("\n✅ Computed output matches the expected output in memory.")
else:
    print("\n❌ Computed output does NOT match the expected output in memory.")

# Optional: Write back computed output to memory
# insert_matrix_into_memory(memory_data, Computed_Output, OUT_start_row, OUT_start_col, mem_cols)
# write_hex_file(file_path, memory_data)
