import numpy as np

# ========== User Configurable Parameters ==========

INPUT_FILE_PATH = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem1.hex"
TOTAL_ROWS_INPUT = 16
TOTAL_COLS_INPUT = 12

A_START_ROW = 0
A_START_COL = 0
A_ROWS = 16
A_COLS = 12

OUTPUT_FILE_PATH = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"
TOTAL_ROWS_OUTPUT = 20
TOTAL_COLS_OUTPUT = 80

OUT_START_ROW = 0
OUT_START_COL = 0
OUT_ROWS = 16
OUT_COLS = 3  # number of sparse vectors generated

S = 2  # sparsity param

NUM_VECTORS = 3  # <-- number of full-length sparse vectors to generate

def read_memory_hex(file_path, rows, cols):
    matrix = np.zeros((rows, cols), dtype=np.int16)
    with open(file_path, 'r') as f:
        lines = f.read().strip().split('\n')
        for idx, line in enumerate(lines):
            r = idx // cols
            c = idx % cols
            if r < rows:
                 val = int(line, 16)
                 # Convert unsigned to signed 16-bit representation:
                 if val >= 0x8000:
                    val -= 0x10000
                 matrix[r, c] = val
    return matrix

def generate_sparse_random_vector(length, S=2):
    vec = np.zeros(length, dtype=np.int8)
    for i in range(length):
        rand_val = np.random.randint(0, 65536)
        if rand_val < (65536 // S):
            vec[i] = 1 if (rand_val & 1) == 0 else -1
    return vec

def compare_matrices(mat1, mat2):
    diff = mat1.astype(np.int32) - mat2.astype(np.int32)
    abs_diff = np.abs(diff)
    max_err = np.max(abs_diff)
    mean_err = np.mean(abs_diff)
    num_errors = np.count_nonzero(diff)

    print("\n--- Comparison Results ---")
    print(f"Max error: {max_err}")
    print(f"Mean error: {mean_err:.4f}")
    print(f"Number of differing elements: {num_errors} / {mat1.size}")

    if num_errors > 0:
        print("\nPositions with errors (row, col): difference")
        rows, cols = np.where(diff != 0)
        for r, c in zip(rows, cols):
            print(f"({r}, {c}): {diff[r,c]}")

    return diff

def main():
    # Load input and output matrices
    input_matrix = read_memory_hex(INPUT_FILE_PATH, TOTAL_ROWS_INPUT, TOTAL_COLS_INPUT)
    A_block = input_matrix[A_START_ROW:A_START_ROW+A_ROWS, A_START_COL:A_START_COL+A_COLS]
    print("Input matrix block (A):")
    print(A_block)

    output_matrix_mem = read_memory_hex(OUTPUT_FILE_PATH, TOTAL_ROWS_OUTPUT, TOTAL_COLS_OUTPUT)
    output_block_expected = output_matrix_mem[OUT_START_ROW:OUT_START_ROW+OUT_ROWS, OUT_START_COL:OUT_START_COL+OUT_COLS]
    print("\nExpected output matrix block:")
    print(output_block_expected)

    np.random.seed(42)  # reproducibility

    # Generate NUM_VECTORS full-length sparse vectors and multiply
    sparse_vectors = []
    output_matrix_calc = np.zeros((A_ROWS, NUM_VECTORS), dtype=np.int16)

    for v in range(NUM_VECTORS):
        vec = generate_sparse_random_vector(A_COLS, S)
        sparse_vectors.append(vec)
        print(f"\nSparse vector {v}:")
        print(vec)

        # Multiply A_block (A_ROWS x A_COLS) by vec (A_COLS x 1) -> result (A_ROWS x 1)
        result_col = A_block.dot(vec)
        output_matrix_calc[:, v] = result_col

    print("\nCalculated output matrix:")
    print(output_matrix_calc)

    if output_block_expected.shape == output_matrix_calc.shape:
        diff = compare_matrices(output_block_expected, output_matrix_calc)
        print("\nDifference matrix:")
        print(diff)
    else:
        print("\nWarning: Output block shape and calculated output shape differ! Cannot compare.")

if __name__ == "__main__":
    main()
